"""
@file
@brief Implements a class able to compute the predictions
from on an :epkg:`ONNX` model.
"""
from onnx import helper, shape_inference
from .onnx2py_helper import guess_proto_dtype


def enumerate_model_node_outputs(model, add_node=False):
    """
    Enumerates all the nodes of a model.

    @param      model       :epkg:`ONNX` graph
    @param      add_node    if False, the function enumerates
                            all output names from every node, otherwise, it
                            enumerates tuple (output name, node)
    @return                 enumerator
    """
    if not hasattr(model, "graph"):
        raise TypeError(  # pragma: no cover
            "Parameter model is not an ONNX model but "
            "{}".format(type(model)))
    for node in model.graph.node:
        for out in node.output:
            yield (out, node) if add_node else out


def select_model_inputs_outputs(model, outputs=None, inputs=None,
                                infer_shapes=False, overwrite=None,
                                verbose=0, fLOG=None):
    """
    Takes a model and changes its outputs.

    :param model: :epkg:`ONNX` model
    :param inputs: new inputs, same ones if None
    :param outputs: new outputs, same ones if None
    :param infer_shapes: infer inputs and outputs shapes
    :param overwrite: overwrite type and shapes for
        inputs or outputs, *overwrite* is a
        dictionary `{'name': (numpy dtype, shape)}`
    :param verbose: display information while converting
    :param fLOG: logging function
    :return: modified model

    The function removes unneeded nodes.

    .. exref::
        :title: Change ONNX model inputs

        The following exampels shows how to change the inputs of model
        to bypass the first nodes. Shape inferences fails to determine
        the new inputs type. They need to be overwritten.
        `verbose=1, fLOG=print` shows the number of deleted nodes.

        ::

            import onnx
            from mlprodict.tools.onnx_manipulations import select_model_inputs_outputs

            onx = onnx.load(path)
            onx2 = select_model_inputs_outputs(
                onx, inputs=["SentenceTokenizer/SentencepieceTokenizeOp:0",
                             "SentenceTokenizer/SentencepieceTokenizeOp:1"],
                infer_shapes=True, verbose=1, fLOG=print,
                overwrite={'SentenceTokenizer/SentencepieceTokenizeOp:0': (numpy.int32, None),
                           'SentenceTokenizer/SentencepieceTokenizeOp:1': (numpy.int64, None)})
            onnx.save(onx2, path2)

    .. versionchanged:: 0.6
        Supports the case where inputs are changed.
    """
    if inputs is not None and not isinstance(inputs, list):
        inputs = [inputs]
    if outputs is not None and not isinstance(outputs, list):
        outputs = [outputs]
    if inputs is None:
        inputs = [i.name for i in model.graph.input]
    if outputs is None:
        outputs = [o.name for o in model.graph.output]

    mark_var = {}
    for out in enumerate_model_node_outputs(model):
        mark_var[out] = 0
    for inp in inputs:
        mark_var[inp] = 0
    for out in outputs:
        if out not in mark_var:
            raise ValueError(  # pragma: no cover
                "Output '{}' not found in model.".format(out))
        mark_var[out] = 1

    nodes = model.graph.node[::-1]
    mark_op = {}
    for node in nodes:
        mark_op[node.name] = 0

    # We mark all the nodes we need to keep.
    nb = 1
    while nb > 0:
        nb = 0
        for node in nodes:
            if mark_op[node.name] == 1:
                continue
            mod = False
            for out in node.output:
                if mark_var[out] == 1:
                    mark_op[node.name] = 1
                    mod = True
                    break
            if not mod:
                continue

            nb += 1
            for inp in node.input:
                if inp in inputs:
                    continue
                if mark_var.get(inp, 0) == 1:
                    continue
                mark_var[inp] = 1
                nb += 1

    # All nodes verifies mark_op[node.name] == 1
    keep_nodes = [node for node in nodes if mark_op[node.name] == 1]

    known_shapes = {}
    if infer_shapes:
        shapes = shape_inference.infer_shapes(model)
        for shape in shapes.graph.value_info:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type
        for shape in shapes.graph.input:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type
        for shape in shapes.graph.output:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type
    else:
        for shape in model.graph.input:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type
        for shape in model.graph.output:  # pylint: disable=E1101
            known_shapes[shape.name] = shape.type

    var_in = []
    for name in inputs:
        if overwrite is not None and name in overwrite:
            dtype, shape = overwrite[name]
            proto_dtype = guess_proto_dtype(dtype)
            value_info = helper.make_tensor_value_info(
                name, proto_dtype, shape)
        elif name in known_shapes:
            info = known_shapes[name].tensor_type
            proto_dtype = info.elem_type
            if proto_dtype == 0:
                value_info = helper.ValueInfoProto()
                value_info.name = name
            else:
                shape = [getattr(d, 'dim_value', None) for d in info.shape.dim]
                shape = [None if s == 0 else s for s in shape]
                value_info = helper.make_tensor_value_info(
                    name, proto_dtype, shape)
        else:
            value_info = helper.ValueInfoProto()
            value_info.name = name
        var_in.append(value_info)

    var_out = []
    for name in outputs:
        if overwrite is not None and name in overwrite:
            dtype, shape = overwrite[name]
            proto_dtype = guess_proto_dtype(dtype)
            value_info = helper.make_tensor_value_info(
                name, proto_dtype, shape)
        elif name in known_shapes:
            info = known_shapes[name].tensor_type
            proto_dtype = info.elem_type
            if proto_dtype == 0:
                value_info = helper.ValueInfoProto()
                value_info.name = name
            else:
                shape = [getattr(d, 'dim_value', None) for d in info.shape.dim]
                shape = [None if s == 0 else s for s in shape]
                value_info = helper.make_tensor_value_info(
                    name, proto_dtype, shape)
        else:
            value_info = helper.ValueInfoProto()
            value_info.name = name
        var_out.append(value_info)

    if verbose > 0 and fLOG is not None:  # pragma: no cover
        fLOG("[select_model_inputs_outputs] nodes %r --> %r" % (
            len(model.graph.node), len(keep_nodes)))
        fLOG("[select_model_inputs_outputs] inputs: %r" % var_in)
        fLOG("[select_model_inputs_outputs] inputs: %r" % var_out)

    graph = helper.make_graph(keep_nodes, model.graph.name, var_in,
                              var_out, model.graph.initializer)
    onnx_model = helper.make_model(graph)
    onnx_model.ir_version = model.ir_version
    onnx_model.producer_name = model.producer_name
    onnx_model.producer_version = model.producer_version
    onnx_model.domain = model.domain
    onnx_model.model_version = model.model_version
    onnx_model.doc_string = model.doc_string
    if len(model.metadata_props) > 0:  # pragma: no cover
        values = {p.key: p.value for p in model.metadata_props}
        helper.set_model_props(onnx_model, values)

    del onnx_model.opset_import[:]  # pylint: disable=E1101
    for oimp in model.opset_import:
        op_set = onnx_model.opset_import.add()  # pylint: disable=E1101
        op_set.domain = oimp.domain
        op_set.version = oimp.version
    return onnx_model



"""
@file
@brief Tools to test models from the :epkg:`ONNX Zoo`.

.. versionadded:: 0.6
"""
import os
import urllib.request
from collections import OrderedDict
import numpy
from onnx import TensorProto, numpy_helper


def short_list_zoo_models():
    """
    Returns a short list from :epkg:`ONNX Zoo`.

    :return: list of dictionaries.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import pprint
        from mlprodict.tools.zoo import short_list_zoo_models
        pprint.pprint(short_list_zoo_models())
    """
    return [
        dict(name="mobilenet",
             model="https://github.com/onnx/models/raw/master/vision/"
                   "classification/mobilenet/model/mobilenetv2-7.tar.gz"),
        dict(name="resnet18",
             model="https://github.com/onnx/models/raw/master/vision/"
                   "classification/resnet/model/resnet18-v1-7.tar.gz"),
        dict(name="squeezenet",
             model="https://github.com/onnx/models/raw/master/vision/"
                   "classification/squeezenet/model/squeezenet1.0-9.tar.gz",
             folder="squeezenet"),
        dict(name="densenet121",
             model="https://github.com/onnx/models/raw/master/vision/"
                   "classification/densenet-121/model/densenet-9.tar.gz",
             folder="densenet121"),
        dict(name="inception2",
             model="https://github.com/onnx/models/raw/master/vision/"
                   "classification/inception_and_googlenet/inception_v2/"
                   "model/inception-v2-9.tar.gz"),
        dict(name="shufflenet",
             model="https://github.com/onnx/models/raw/master/vision/"
                   "classification/shufflenet/model/shufflenet-9.tar.gz"),
        dict(name="efficientnet-lite4",
             model="https://github.com/onnx/models/raw/master/vision/"
                   "classification/efficientnet-lite4/model/"
                   "efficientnet-lite4-11.tar.gz"),
    ]


def _download_url(url, output_path, name, verbose=False):
    if verbose:  # pragma: no cover
        from tqdm import tqdm

        class DownloadProgressBar(tqdm):
            "progress bar hook"

            def update_to(self, b=1, bsize=1, tsize=None):
                "progress bar hook"
                if tsize is not None:
                    self.total = tsize
                self.update(b * bsize - self.n)

        with DownloadProgressBar(unit='B', unit_scale=True,
                                 miniters=1, desc=name) as t:
            urllib.request.urlretrieve(
                url, filename=output_path, reporthook=t.update_to)
    else:
        urllib.request.urlretrieve(url, filename=output_path)


def load_data(folder):
    """
    Restores protobuf data stored in a folder.

    :param folder: folder
    :return: dictionary
    """
    res = OrderedDict()
    res['in'] = OrderedDict()
    res['out'] = OrderedDict()
    files = os.listdir(folder)
    for name in files:
        noext, ext = os.path.splitext(name)
        if ext == '.pb':
            data = TensorProto()
            with open(os.path.join(folder, name), 'rb') as f:
                data.ParseFromString(f.read())
            if noext.startswith('input'):
                res['in'][noext] = numpy_helper.to_array(data)
            elif noext.startswith('output'):
                res['out'][noext] = numpy_helper.to_array(data)
            else:
                raise ValueError(  # pragma: no cover
                    "Unable to guess anything about %r." % noext)

    return res


def download_model_data(name, model=None, cache=None, verbose=False):
    """
    Downloads a model and returns a link to the local
    :epkg:`ONNX` file and data which can be used as inputs.

    :param name: model name (see @see fn short_list_zoo_models)
    :param model: url or empty to get the default value
        returned by @see fn short_list_zoo_models)
    :param cache: folder to cache the downloaded data
    :param verbose: display a progress bar
    :return: local onnx file, input data
    """
    suggested_folder = None
    if model is None:
        model_list = short_list_zoo_models()
        for mod in model_list:
            if mod['name'] == name:
                model = mod['model']
                if 'folder' in mod:  # pylint: disable=R1715
                    suggested_folder = mod['folder']
                break
        if model is None:
            raise ValueError(
                "Unable to find a default value for name=%r." % name)

    # downloads
    last_name = model.split('/')[-1]
    if cache is None:
        cache = os.path.abspath('.')  # pragma: no cover
    dest = os.path.join(cache, last_name)
    if not os.path.exists(dest):
        _download_url(model, dest, name, verbose=verbose)
    size = os.stat(dest).st_size
    if size < 2 ** 20:  # pragma: no cover
        os.remove(dest)
        raise ConnectionError(
            "Unable to download model from %r." % model)

    outtar = os.path.splitext(dest)[0]
    if not os.path.exists(outtar):
        from pyquickhelper.filehelper.compression_helper import (
            ungzip_files)
        ungzip_files(dest, unzip=False, where_to=cache, remove_space=False)

    onnx_file = os.path.splitext(outtar)[0]
    if not os.path.exists(onnx_file):
        from pyquickhelper.filehelper.compression_helper import (
            untar_files)
        untar_files(outtar, where_to=cache)

    if suggested_folder is not None:
        fold_onnx = [suggested_folder]
    else:
        fold_onnx = [onnx_file, onnx_file.split('-')[0],
                     '-'.join(onnx_file.split('-')[:-1]),
                     '-'.join(onnx_file.split('-')[:-1]).replace('-', '_')]
    fold_onnx_ok = [_ for _ in fold_onnx if os.path.exists(_)]
    if len(fold_onnx_ok) != 1:
        raise FileNotFoundError(  # pragma: no cover
            "Unable to find an existing folder among %r." % fold_onnx)
    onnx_file = fold_onnx_ok[0]

    onnx_files = [_ for _ in os.listdir(onnx_file) if _.endswith(".onnx")]
    if len(onnx_files) != 1:
        raise FileNotFoundError(  # pragma: no cover
            "Unable to find any onnx file in %r." % onnx_files)
    final_onnx = os.path.join(onnx_file, onnx_files[0])

    # data
    data = [_ for _ in os.listdir(onnx_file)
            if os.path.isdir(os.path.join(onnx_file, _))]
    examples = OrderedDict()
    for f in data:
        examples[f] = load_data(os.path.join(onnx_file, f))

    return final_onnx, examples


def verify_model(onnx_file, examples, runtime=None, abs_tol=5e-4,
                 verbose=0, fLOG=None):
    """
    Verifies a model.

    :param onnx_file: ONNX file
    :param examples: list of examples to verify
    :param runtime: a runtime to use
    :param abs_tol: error tolerance when checking the output
    :param verbose: verbosity level for for runtime other than
        `'onnxruntime'`
    :param fLOG: logging function when `verbose > 0`
    :return: errors for every sample
    """
    if runtime == 'onnxruntime':
        from onnxruntime import InferenceSession
        sess = InferenceSession(onnx_file)
        meth = lambda data, s=sess: s.run(None, data)
        names = [p.name for p in sess.get_inputs()]
        onames = list(range(len(sess.get_outputs())))
    else:
        def _lin_(sess, data, names):
            r = sess.run(data, verbose=verbose, fLOG=fLOG)
            return [r[n] for n in names]

        from ..onnxrt import OnnxInference
        sess = OnnxInference(onnx_file, runtime=runtime)
        names = sess.input_names
        onames = sess.output_names
        meth = lambda data, s=sess, ns=onames: _lin_(s, data, ns)

    rows = []
    for index, (name, data_inout) in enumerate(examples.items()):
        data = data_inout["in"]
        if len(data) != len(names):
            raise RuntimeError(  # pragma: no cover
                "Mismathed number of inputs %d != %d\ninputs: %r\nmodel: %r."
                "" % (len(data), len(names), list(sorted(data)), names))
        inputs = {n: data[v] for n, v in zip(names, data)}
        outputs = meth(inputs)
        expected = data_inout['out']
        if len(outputs) != len(onames):
            raise RuntimeError(  # pragma: no cover
                "Number of outputs %d is != expected outputs %d." % (
                    len(outputs), len(onames)))
        for i, (output, expect) in enumerate(zip(outputs, expected.items())):
            if output.shape != expect[1].shape:
                raise ValueError(  # pragma: no cover
                    "Shape mismatch got %r != expected %r." % (
                        output.shape, expect[1].shape))
            diff = numpy.abs(output - expect[1]).ravel()
            absolute = diff.max()
            relative = absolute / numpy.median(diff) if absolute > 0 else 0.
            if absolute > abs_tol:
                raise ValueError(  # pragma: no cover
                    "Example %d, inferred and expected resuls are different "
                    "for output %d: abs=%r rel=%r (runtime=%r)."
                    "" % (index, i, absolute, relative, runtime))
            rows.append(dict(name=name, i=i, abs=absolute, rel=relative))
    return rows



"""
@file
@brief Statistics on :epkg:`ONNX` models.
"""
from collections import Counter
from onnx.helper import make_graph
from onnx import ValueInfoProto
from skl2onnx.common._topology import Variable
from ._onnx_optimisation_common import _apply_optimisation_on_graph
from .onnx_optimisation import onnx_remove_node


def onnx_statistics(onnx_model, recursive=True, optim=True):
    """
    Computes statistics on :epkg:`ONNX` models,
    extracts informations about the model such as
    the number of nodes.

    @param      onnx_model      onnx model
    @param      recursive       looks into subgraphs
    @param      optim           adds statistics because of optimisation
    @return                     dictionary

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import pprint
        from sklearn.linear_model import LogisticRegression
        from sklearn.ensemble import RandomForestClassifier
        from sklearn.datasets import load_iris
        from mlprodict.onnxrt.optim.onnx_helper import onnx_statistics
        from mlprodict.onnx_conv import to_onnx

        iris = load_iris()
        X = iris.data
        y = iris.target
        lr = LogisticRegression()
        lr.fit(X, y)
        onx = to_onnx(lr, X[:1])
        pprint.pprint((lr, onnx_statistics(onx)))

        iris = load_iris()
        X = iris.data
        y = iris.target
        rf = RandomForestClassifier()
        rf.fit(X, y)
        onx = to_onnx(rf, X[:1], target_opset=12)
        pprint.pprint((rf, onnx_statistics(onx)))
    """
    atts = ['doc_string', 'ir_version', 'metadata_props', 'domain',
            'model_version', 'producer_name', 'producer_version']

    def update(sts, st):
        for k, v in st.items():
            if k in ['size'] or k in atts:
                continue  # pragma: no cover
            if k in sts:
                sts[k] += v
            else:
                sts[k] = v

    if hasattr(onnx_model, 'graph'):
        content = onnx_model.SerializeToString()
        nnodes = len(onnx_model.graph.node)
        ninits = len(onnx_model.graph.initializer)
        stats = {'size': len(content), 'nnodes': nnodes, 'ninits': ninits}
        for a in atts:
            v = getattr(onnx_model, a)
            if isinstance(v, str):
                li = None
            else:
                try:
                    li = list(v)
                except TypeError:
                    li = None
            if li is not None and len(li) == 0:
                continue
            stats[a] = v

        for opi in onnx_model.opset_import:
            stats[opi.domain] = opi.version

        graph = onnx_model.graph
    elif not hasattr(onnx_model, 'node'):  # pragma: no cover
        # We're in a node.
        stats = {'nnodes': 1}
        if hasattr(onnx_model, 'attribute') and onnx_model.attribute:
            for att in onnx_model.attribute:
                if att.name == 'body':
                    st = onnx_statistics(att.g)
                    update(stats, st)
        return stats
    else:
        graph = onnx_model
        nnodes = len(graph.node)
        stats = {'nnodes': nnodes}

    # Number of identities
    counts = Counter(map(lambda obj: obj.op_type, graph.node))
    for op in ['Cast', 'Identity', 'ZipMap', 'Reshape']:
        if op in counts:
            stats['op_' + op] = counts[op]

    # Recursive
    if recursive:
        for node in graph.node:
            if not hasattr(node, 'attribute'):
                continue  # pragma: no cover
            for att in node.attribute:
                if att.name != 'body':
                    continue
                substats = onnx_statistics(att.g, recursive=True, optim=False)
                update(stats, {'subgraphs': 1})
                update(stats, substats)

    # optimisation: remove_identity nodes
    if optim:
        new_model = onnx_remove_node(
            onnx_model, recursive=recursive)
        st = onnx_statistics(new_model, recursive=recursive, optim=False)
        for key in ["op_Identity", "subgraphs", "size",
                    "nnodes", "ninits"]:
            if key in st:
                stats[key + "_optim"] = st[key]
    return stats


def change_input_first_dimension(onnx_model, N=None, debug_info=None):
    """
    Some models are converted under the assumption
    batch prediction is not necessary. This function
    changes the first dimension of an ONNX graph.

    @param      onnx_model      model :epkg:`onnx`
    @param      N               new first dimension,
                                None to avoid changing it,
                                0 to fix an undefined
                                first dimension
    @param      debug_info      unused
    @return                     modified model onnx
    """
    def _make_value_info(variable):
        value_info = ValueInfoProto()
        value_info.name = variable.full_name
        value_info.type.CopyFrom(  # pylint: disable=E1101
            variable.type.to_onnx_type())  # pylint: disable=E1101
        if variable.type.doc_string:  # pylint: disable=E0611
            value_info.doc_string = variable.type.doc_string  # pragma: no cover
        return value_info

    if hasattr(onnx_model, 'graph'):
        return _apply_optimisation_on_graph(
            change_input_first_dimension, onnx_model, N=N)

    graph = onnx_model

    nodes = graph.node
    inputs = [Variable.from_pb(input) for input in onnx_model.input]
    outputs = onnx_model.output

    if N <= 0:
        N = None
    for input in inputs:
        input.type.shape[0] = N
    inputs = [_make_value_info(v) for v in inputs]

    graph = make_graph(nodes, onnx_model.name,
                       inputs, outputs, onnx_model.initializer)

    graph.value_info.extend(onnx_model.value_info)  # pylint: disable=E1101
    return graph



"""
@file
@brief Extensions to class @see cl OnnxInference.
"""
import os
import json
import re
from io import BytesIO
import pickle
import textwrap
from onnx import numpy_helper
from ..tools.onnx2py_helper import _var_as_dict, _type_to_string


class OnnxInferenceExport:
    """
    Implements methods to export a instance of
    @see cl OnnxInference into :epkg:`json` or :epkg:`dot`.
    """

    def __init__(self, oinf):
        """
        @param      oinf    @see cl OnnxInference
        """
        self.oinf = oinf

    def to_dot(self, recursive=False, prefix='',  # pylint: disable=R0914
               add_rt_shapes=False, use_onnx=False, **params):
        """
        Produces a :epkg:`DOT` language string for the graph.

        :param params: additional params to draw the graph
        :param recursive: also show subgraphs inside operator like
            @see cl Scan
        :param prefix: prefix for every node name
        :param add_rt_shapes: adds shapes infered from the python runtime
        :param use_onnx: use :epkg:`onnx` dot format instead of this one
        :return: string

        Default options for the graph are:

        ::

            options = {
                'orientation': 'portrait',
                'ranksep': '0.25',
                'nodesep': '0.05',
                'width': '0.5',
                'height': '0.1',
            }

        One example:

        .. exref::
            :title: Convert ONNX into DOT

            An example on how to convert an :epkg:`ONNX`
            graph into :epkg:`DOT`.

            .. runpython::
                :showcode:
                :warningout: DeprecationWarning

                import numpy
                from skl2onnx.algebra.onnx_ops import OnnxLinearRegressor
                from skl2onnx.common.data_types import FloatTensorType
                from mlprodict.onnxrt import OnnxInference

                pars = dict(coefficients=numpy.array([1., 2.]),
                            intercepts=numpy.array([1.]),
                            post_transform='NONE')
                onx = OnnxLinearRegressor('X', output_names=['Y'], **pars)
                model_def = onx.to_onnx({'X': pars['coefficients'].astype(numpy.float32)},
                                        outputs=[('Y', FloatTensorType([1]))],
                                        target_opset=12)
                oinf = OnnxInference(model_def)
                print(oinf.to_dot())

            See an example of representation in notebook
            :ref:`onnxvisualizationrst`.
        """
        clean_label_reg1 = re.compile("\\\\x\\{[0-9A-F]{1,6}\\}")
        clean_label_reg2 = re.compile("\\\\p\\{[0-9P]{1,6}\\}")

        def dot_name(text):
            return text.replace("/", "_").replace(":", "__")

        def dot_label(text):
            for reg in [clean_label_reg1, clean_label_reg2]:
                fall = reg.findall(text)
                for f in fall:
                    text = text.replace(f, "_")  # pragma: no cover
            return text

        options = {
            'orientation': 'portrait',
            'ranksep': '0.25',
            'nodesep': '0.05',
            'width': '0.5',
            'height': '0.1',
        }
        options.update(params)

        if use_onnx:
            from onnx.tools.net_drawer import GetPydotGraph, GetOpNodeProducer

            pydot_graph = GetPydotGraph(
                self.oinf.obj.graph, name=self.oinf.obj.graph.name,
                rankdir=params.get('rankdir', "TB"),
                node_producer=GetOpNodeProducer(
                    "docstring", fillcolor="orange", style="filled",
                    shape="box"))
            return pydot_graph.to_string()

        inter_vars = {}
        exp = ["digraph{"]
        for opt in {'orientation', 'pad', 'nodesep', 'ranksep'}:
            if opt in options:
                exp.append("  {}={};".format(opt, options[opt]))
        fontsize = 10

        shapes = {}
        if add_rt_shapes:
            if not hasattr(self.oinf, 'shapes_'):
                raise RuntimeError(  # pragma: no cover
                    "No information on shapes, check the runtime '{}'.".format(self.oinf.runtime))
            for name, shape in self.oinf.shapes_.items():
                va = shape.evaluate().to_string()
                shapes[name] = va
                if name in self.oinf.inplaces_:
                    shapes[name] += "\\ninplace"

        # inputs
        exp.append("")
        for obj in self.oinf.obj.graph.input:
            dobj = _var_as_dict(obj)
            sh = shapes.get(dobj['name'], '')
            if sh:
                sh = "\\nshape={}".format(sh)
            exp.append(
                '  {3}{0} [shape=box color=red label="{0}\\n{1}{4}" fontsize={2}];'.format(
                    dot_name(dobj['name']), _type_to_string(dobj['type']),
                    fontsize, prefix, dot_label(sh)))
            inter_vars[obj.name] = obj

        # outputs
        exp.append("")
        for obj in self.oinf.obj.graph.output:
            dobj = _var_as_dict(obj)
            sh = shapes.get(dobj['name'], '')
            if sh:
                sh = "\\nshape={}".format(sh)
            exp.append(
                '  {3}{0} [shape=box color=green label="{0}\\n{1}{4}" fontsize={2}];'.format(
                    dot_name(dobj['name']), _type_to_string(dobj['type']),
                    fontsize, prefix, dot_label(sh)))
            inter_vars[obj.name] = obj

        # initializer
        exp.append("")
        for obj in self.oinf.obj.graph.initializer:
            dobj = _var_as_dict(obj)
            val = dobj['value']
            flat = val.flatten()
            if flat.shape[0] < 9:
                st = str(val)
            else:
                st = str(val)
                if len(st) > 50:
                    st = st[:50] + '...'
            st = st.replace('\n', '\\n')
            kind = ""
            exp.append(
                '  {6}{0} [shape=box label="{0}\\n{4}{1}({2})\\n{3}" fontsize={5}];'.format(
                    dot_name(dobj['name']), dobj['value'].dtype,
                    dobj['value'].shape, dot_label(st), kind, fontsize, prefix))
            inter_vars[obj.name] = obj

        # nodes
        fill_names = {}
        for node in self.oinf.obj.graph.node:
            exp.append("")
            for out in node.output:
                if out not in inter_vars:
                    inter_vars[out] = out
                    sh = shapes.get(out, '')
                    if sh:
                        sh = "\\nshape={}".format(sh)
                    exp.append(
                        '  {2}{0} [shape=box label="{0}{3}" fontsize={1}];'.format(
                            dot_name(out), fontsize, dot_name(prefix), dot_label(sh)))

            dobj = _var_as_dict(node)
            if dobj['name'].strip() == '':  # pragma: no cover
                name = node.op_type
                iname = 1
                while name in fill_names:
                    name = "%s%d" % (name, iname)
                    iname += 1
                dobj['name'] = name
                node.name = name

            atts = []
            if 'atts' in dobj:
                for k, v in sorted(dobj['atts'].items()):
                    val = None
                    if 'value' in v:
                        val = str(v['value']).replace(
                            "\n", "\\n").replace('"', "'")
                        sl = max(30 - len(k), 10)
                        if len(val) > sl:
                            val = val[:sl] + "..."
                    if val is not None:
                        atts.append('{}={}'.format(k, val))
            satts = "" if len(atts) == 0 else ("\\n" + "\\n".join(atts))

            if recursive and node.op_type in {'Scan', 'Loop', 'If'}:
                fields = (['then_branch', 'else_branch']
                          if node.op_type == 'If' else ['body'])
                for field in fields:
                    if field not in dobj['atts']:
                        continue  # pragma: no cover

                    # creates the subgraph
                    body = dobj['atts'][field]['value']
                    oinf = self.oinf.__class__(
                        body, runtime=self.oinf.runtime, skip_run=self.oinf.skip_run)
                    subprefix = prefix + "B_"
                    subdot = oinf.to_dot(recursive=recursive, prefix=subprefix,
                                         add_rt_shapes=add_rt_shapes)
                    lines = subdot.split("\n")
                    start = 0
                    for i, line in enumerate(lines):
                        if '[' in line:
                            start = i
                            break
                    subgraph = "\n".join(lines[start:])

                    # connecting the subgraph
                    exp.append("  subgraph cluster_{}{} {{".format(
                        node.op_type, id(node)))
                    exp.append('    label="{0}\\n({1}){2}";'.format(
                        dobj['op_type'], dot_name(dobj['name']), satts))
                    exp.append('    fontsize={0};'.format(fontsize))
                    exp.append('    color=black;')
                    exp.append(
                        '\n'.join(map(lambda s: '  ' + s, subgraph.split('\n'))))

                    for inp1, inp2 in zip(node.input, body.input):
                        exp.append(
                            "  {0}{1} -> {2}{3};".format(
                                dot_name(prefix), dot_name(inp1),
                                dot_name(subprefix), dot_name(inp2.name)))
                    for out1, out2 in zip(body.output, node.output):
                        exp.append(
                            "  {0}{1} -> {2}{3};".format(
                                dot_name(subprefix), dot_name(out1.name),
                                dot_name(prefix), dot_name(out2)))

            else:
                exp.append('  {4}{1} [shape=box style="filled,rounded" color=orange label="{0}\\n({1}){2}" fontsize={3}];'.format(
                    dobj['op_type'], dot_name(dobj['name']), satts, fontsize,
                    dot_name(prefix)))

                for inp in node.input:
                    exp.append(
                        "  {0}{1} -> {0}{2};".format(
                            dot_name(prefix), dot_name(inp), dot_name(node.name)))
                for out in node.output:
                    exp.append(
                        "  {0}{1} -> {0}{2};".format(
                            dot_name(prefix), dot_name(node.name), dot_name(out)))

        exp.append('}')
        return "\n".join(exp)

    def to_json(self, indent=2):
        """
        Converts an :epkg:`ONNX` model into :epkg:`JSON`.

        @param      indent      indentation
        @return                 string

        .. exref::
            :title: Convert ONNX into JSON

            An example on how to convert an :epkg:`ONNX`
            graph into :epkg:`JSON`.

            .. runpython::
                :showcode:
                :warningout: DeprecationWarning

                import numpy
                from skl2onnx.algebra.onnx_ops import OnnxLinearRegressor
                from skl2onnx.common.data_types import FloatTensorType
                from mlprodict.onnxrt import OnnxInference

                pars = dict(coefficients=numpy.array([1., 2.]),
                            intercepts=numpy.array([1.]),
                            post_transform='NONE')
                onx = OnnxLinearRegressor('X', output_names=['Y'], **pars)
                model_def = onx.to_onnx({'X': pars['coefficients'].astype(numpy.float32)},
                                        outputs=[('Y', FloatTensorType([1]))],
                                        target_opset=12)
                oinf = OnnxInference(model_def)
                print(oinf.to_json())
        """

        def _to_json(obj):
            s = str(obj)
            rows = ['{']
            leave = None
            for line in s.split('\n'):
                if line.endswith("{"):
                    rows.append('"%s": {' % line.strip('{ '))
                elif ':' in line:
                    spl = line.strip().split(':')
                    if len(spl) != 2:
                        raise RuntimeError(  # pragma: no cover
                            "Unable to interpret line '{}'.".format(line))

                    if spl[0].strip() in ('type', ):
                        st = spl[1].strip()
                        if st in {'INT', 'INTS', 'FLOAT', 'FLOATS',
                                  'STRING', 'STRINGS', 'TENSOR'}:
                            spl[1] = '"{}"'.format(st)

                    if spl[0] in ('floats', 'ints'):
                        if leave:
                            rows.append("{},".format(spl[1]))
                        else:
                            rows.append('"{}": [{},'.format(
                                spl[0], spl[1].strip()))
                            leave = spl[0]
                    elif leave:
                        rows[-1] = rows[-1].strip(',')
                        rows.append('],')
                        rows.append('"{}": {},'.format(
                            spl[0].strip(), spl[1].strip()))
                        leave = None
                    else:
                        rows.append('"{}": {},'.format(
                            spl[0].strip(), spl[1].strip()))
                elif line.strip() == "}":
                    rows[-1] = rows[-1].rstrip(",")
                    rows.append(line + ",")
                elif line:
                    raise RuntimeError(  # pragma: no cover
                        "Unable to interpret line '{}'.".format(line))
            rows[-1] = rows[-1].rstrip(',')
            rows.append("}")
            js = "\n".join(rows)

            try:
                content = json.loads(js)
            except json.decoder.JSONDecodeError as e:  # pragma: no cover
                js2 = "\n".join("%04d %s" % (i + 1, line)
                                for i, line in enumerate(js.split("\n")))
                raise RuntimeError(
                    "Unable to parse JSON\n{}".format(js2)) from e
            return content

        # meta data
        final_obj = {}
        for k in {'ir_version', 'producer_name', 'producer_version',
                  'domain', 'model_version', 'doc_string'}:
            if hasattr(self.oinf.obj, k):
                final_obj[k] = getattr(self.oinf.obj, k)

        # inputs
        inputs = []
        for obj in self.oinf.obj.graph.input:
            st = _to_json(obj)
            inputs.append(st)
        final_obj['inputs'] = inputs

        # outputs
        outputs = []
        for obj in self.oinf.obj.graph.output:
            st = _to_json(obj)
            outputs.append(st)
        final_obj['outputs'] = outputs

        # init
        inits = {}
        for obj in self.oinf.obj.graph.initializer:
            value = numpy_helper.to_array(obj).tolist()
            inits[obj.name] = value
        final_obj['initializers'] = inits

        # nodes
        nodes = []
        for obj in self.oinf.obj.graph.node:
            node = dict(name=obj.name, op_type=obj.op_type, domain=obj.domain,
                        inputs=[str(_) for _ in obj.input],
                        outputs=[str(_) for _ in obj.output],
                        attributes={})
            for att in obj.attribute:
                st = _to_json(att)
                node['attributes'][st['name']] = st
                del st['name']
            nodes.append(node)
        final_obj['nodes'] = nodes

        return json.dumps(final_obj, indent=indent)

    def to_python(self, prefix="onnx_pyrt_", dest=None, inline=True):
        """
        Converts the ONNX runtime into independant python code.
        The function creates multiple files starting with
        *prefix* and saved to folder *dest*.

        @param  prefix      file prefix
        @param  dest        destination folder
        @param  inline      constant matrices are put in the python file itself
                            as byte arrays
        @return             file dictionary

        The function does not work if the chosen runtime
        is not *python*.

        .. runpython::
            :showcode:
            :warningout: DeprecationWarning

            import numpy
            from skl2onnx.algebra.onnx_ops import OnnxAdd
            from mlprodict.onnxrt import OnnxInference

            idi = numpy.identity(2)
            onx = OnnxAdd('X', idi, output_names=['Y'],
                          op_version=12)
            model_def = onx.to_onnx({'X': idi.astype(numpy.float32)},
                                    target_opset=12)
            X = numpy.array([[1, 2], [3, 4]], dtype=numpy.float32)
            oinf = OnnxInference(model_def, runtime='python')
            res = oinf.to_python()
            print(res['onnx_pyrt_main.py'])
        """

        def clean_args(args):
            new_args = []
            for v in args:
                # remove python keywords
                if v.startswith('min='):
                    av = 'min_=' + v[4:]
                elif v.startswith('max='):
                    av = 'max_=' + v[4:]
                else:
                    av = v
                new_args.append(av)
            return new_args

        if self.oinf.runtime != 'python':
            raise ValueError(
                "The runtime must be 'python' not '{}'.".format(
                    self.oinf.runtime))

        # metadata
        obj = {}
        for k in {'ir_version', 'producer_name', 'producer_version',
                  'domain', 'model_version', 'doc_string'}:
            if hasattr(self.oinf.obj, k):
                obj[k] = getattr(self.oinf.obj, k)
        code_begin = ["# coding: utf-8",
                      "'''",
                      "Python code equivalent to an ONNX graph.",
                      "It was was generated by module *mlprodict*.",
                      "'''"]
        code_imports = ["from io import BytesIO",
                        "import pickle",
                        "from numpy import array, float32, ndarray"]
        code_lines = ["class OnnxPythonInference:", "",
                      "    def __init__(self):",
                      "        self._load_inits()", "",
                      "    @property",
                      "    def metadata(self):",
                      "        return %r" % obj, ""]

        # inputs
        inputs = [obj.name for obj in self.oinf.obj.graph.input]
        code_lines.extend([
            "    @property", "    def inputs(self):",
            "        return %r" % inputs,
            ""
        ])

        # outputs
        outputs = [obj.name for obj in self.oinf.obj.graph.output]
        code_lines.extend([
            "    @property", "    def outputs(self):",
            "        return %r" % outputs,
            ""
        ])

        # init
        code_lines.extend(["    def _load_inits(self):",
                           "        self._inits = {}"])
        file_data = {}
        for obj in self.oinf.obj.graph.initializer:
            value = numpy_helper.to_array(obj)
            bt = BytesIO()
            pickle.dump(value, bt)
            name = '{1}{0}.pkl'.format(obj.name, prefix)
            if inline:
                code_lines.extend([
                    "        iocst = %r" % bt.getvalue(),
                    "        self._inits['{0}'] = pickle.loads(iocst)".format(
                        obj.name)
                ])
            else:
                file_data[name] = bt.getvalue()
                code_lines.append(
                    "        self._inits['{0}'] = pickle.loads('{1}')".format(
                        obj.name, name))
        code_lines.append('')

        # inputs, outputs
        inputs = self.oinf.input_names

        # nodes
        code_lines.extend(['    def run(self, %s):' % ', '.join(inputs)])
        ops = {}
        code_lines.append('        # constant')
        for obj in self.oinf.obj.graph.initializer:
            code_lines.append(
                "        {0} = self._inits['{0}']".format(obj.name))
        code_lines.append('')
        code_lines.append('        # graph code')
        for node in self.oinf.sequence_:
            fct = 'pyrt_' + node.name
            if fct not in ops:
                ops[fct] = node
            args = []
            args.extend(node.inputs)
            margs = node.modified_args
            if margs is not None:
                args.extend(clean_args(margs))
            code_lines.append("        {0} = {1}({2})".format(
                ', '.join(node.outputs), fct, ', '.join(args)))
        code_lines.append('')
        code_lines.append('        # return')
        code_lines.append('        return %s' % ', '.join(outputs))
        code_lines.append('')

        # operator code
        code_nodes = []
        for name, op in ops.items():
            inputs_args = clean_args(op.inputs_args)

            code_nodes.append('def {0}({1}):'.format(
                name, ', '.join(inputs_args)))
            imps, code = op.to_python(op.python_inputs)
            if imps is not None:
                if not isinstance(imps, list):
                    imps = [imps]
                code_imports.extend(imps)
            code_nodes.append(textwrap.indent(code, '    '))
            code_nodes.extend(['', ''])

        # end
        code_imports = list(sorted(set(code_imports)))
        code_imports.extend(['', ''])
        file_data[prefix + 'main.py'] = "\n".join(
            code_begin + code_imports + code_nodes + code_lines)

        # saves as files
        if dest is not None:
            for k, v in file_data.items():
                ext = os.path.splitext(k)[-1]
                kf = os.path.join(dest, k)
                if ext == '.py':
                    with open(kf, "w", encoding="utf-8") as f:
                        f.write(v)
                elif ext == '.pkl':  # pragma: no cover
                    with open(kf, "wb") as f:
                        f.write(v)
                else:
                    raise NotImplementedError(  # pragma: no cover
                        "Unknown extension for file '{}'.".format(k))
        return file_data



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from onnx.defs import onnx_opset_version
from ._op import OpRun
from ..shape_object import DimensionObject, ShapeObject


class CommonSplit(OpRun):
    """
    Runtime for operator *Split*.
    """

    def __init__(self, onnx_node, desc=None,
                 expected_attributes=None, **options):
        if 'split' not in options:
            options['split'] = None
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=expected_attributes,
                       **options)
        self.nb_outputs = len(onnx_node.output)

    def common_run(self, mat, split):  # pylint: disable=W0221
        if split is None:
            div = mat.shape[self.axis] // self.nb_outputs
            split = [div] * self.nb_outputs
            split[-1] += mat.shape[self.axis] - sum(split)
        sli = [slice(0, s) for s in mat.shape]
        res = []
        pos = 0
        for spl in split:
            sli[self.axis] = slice(pos, pos + spl)
            pos += spl
            res.append(mat[tuple(sli)])
        return tuple(res)

    def common_infer_shapes(self, data, split):  # pylint: disable=W0221
        if split is None:
            return tuple([ShapeObject(None, dtype=data.dtype)
                          for o in range(self.nb_outputs)])
        res = []
        pos = 0
        for spl in split:
            shape = data.copy()
            shape[self.axis] = DimensionObject(spl)
            pos += spl
            res.append(shape)
        return tuple(res)


class Split_2(CommonSplit):
    """
    Runtime for operator *Split*.
    """

    atts = {'axis': 0, 'split': None}

    def __init__(self, onnx_node, desc=None, **options):
        CommonSplit.__init__(self, onnx_node, desc=desc,
                             expected_attributes=Split_2.atts, **options)

    def _run(self, mat):  # pylint: disable=W0221
        return self.common_run(mat, self.split)

    def _infer_shapes(self, data):  # pylint: disable=W0221
        return self.common_infer_shapes(data, self.split)


class Split_11(Split_2):
    """
    Runtime for operator *Split*.
    """
    pass


class Split_13(CommonSplit):
    """
    Runtime for operator *Split*.
    """

    atts = {'axis': 0}

    def __init__(self, onnx_node, desc=None, **options):
        CommonSplit.__init__(self, onnx_node, desc=desc,
                             expected_attributes=Split_13.atts, **options)

    def _run(self, mat, split=None):  # pylint: disable=W0221
        return self.common_run(mat, split)

    def _infer_shapes(self, data, split=None):  # pylint: disable=W0221
        return tuple([ShapeObject(None, dtype=data.dtype)
                      for o in range(self.nb_outputs)])


if onnx_opset_version() >= 13:
    Split = Split_13
elif onnx_opset_version() >= 11:  # pragma: no cover
    Split = Split_11
else:  # pragma: no cover
    Split = Split_2



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from collections import OrderedDict
import numpy
from ._op_helper import _get_typed_class_attribute
from ._op import OpRunClassifierProb, RuntimeTypeError
from ._op_classifier_string import _ClassifierCommon
from ._new_ops import OperatorSchema
from .op_svm_classifier_ import (  # pylint: disable=E0611,E0401
    RuntimeSVMClassifierFloat,
    RuntimeSVMClassifierDouble,
)


class SVMClassifierCommon(OpRunClassifierProb, _ClassifierCommon):

    def __init__(self, dtype, onnx_node, desc=None,
                 expected_attributes=None, **options):
        OpRunClassifierProb.__init__(self, onnx_node, desc=desc,
                                     expected_attributes=expected_attributes,
                                     **options)
        self._init(dtype=dtype)

    def _get_typed_attributes(self, k):
        return _get_typed_class_attribute(self, k, self.__class__.atts)

    def _find_custom_operator_schema(self, op_name):
        """
        Finds a custom operator defined by this runtime.
        """
        if op_name == "SVMClassifierDouble":
            return SVMClassifierDoubleSchema()
        raise RuntimeError(  # pragma: no cover
            "Unable to find a schema for operator '{}'.".format(op_name))

    def _init(self, dtype):
        self._post_process_label_attributes()
        if dtype == numpy.float32:
            self.rt_ = RuntimeSVMClassifierFloat(20)
        elif dtype == numpy.float64:
            self.rt_ = RuntimeSVMClassifierDouble(20)
        else:
            raise RuntimeTypeError(  # pragma: no cover
                "Unsupported dtype={}.".format(dtype))
        atts = [self._get_typed_attributes(k)
                for k in SVMClassifier.atts]
        self.rt_.init(*atts)

    def _run(self, x):  # pylint: disable=W0221
        """
        This is a C++ implementation coming from
        :epkg:`onnxruntime`.
        `svm_classifier.cc
        <https://github.com/microsoft/onnxruntime/blob/master/onnxruntime/core/providers/cpu/ml/svm_classifier.cc>`_.
        See class :class:`RuntimeSVMClassifier
        <mlprodict.onnxrt.ops_cpu.op_svm_classifier_.RuntimeSVMClassifier>`.
        """
        label, scores = self.rt_.compute(x)
        if scores.shape[0] != label.shape[0]:
            scores = scores.reshape(label.shape[0],
                                    scores.shape[0] // label.shape[0])
        return self._post_process_predicted_label(label, scores)


class SVMClassifier(SVMClassifierCommon):

    atts = OrderedDict([
        ('classlabels_ints', numpy.empty(0, dtype=numpy.int64)),
        ('classlabels_strings', []),
        ('coefficients', numpy.empty(0, dtype=numpy.float32)),
        ('kernel_params', numpy.empty(0, dtype=numpy.float32)),
        ('kernel_type', b'NONE'),
        ('post_transform', b'NONE'),
        ('prob_a', numpy.empty(0, dtype=numpy.float32)),
        ('prob_b', numpy.empty(0, dtype=numpy.float32)),
        ('rho', numpy.empty(0, dtype=numpy.float32)),
        ('support_vectors', numpy.empty(0, dtype=numpy.float32)),
        ('vectors_per_class', numpy.empty(0, dtype=numpy.float32)),
    ])

    def __init__(self, onnx_node, desc=None, **options):
        SVMClassifierCommon.__init__(
            self, numpy.float32, onnx_node, desc=desc,
            expected_attributes=SVMClassifier.atts,
            **options)


class SVMClassifierDouble(SVMClassifierCommon):

    atts = OrderedDict([
        ('classlabels_ints', numpy.empty(0, dtype=numpy.int64)),
        ('classlabels_strings', []),
        ('coefficients', numpy.empty(0, dtype=numpy.float64)),
        ('kernel_params', numpy.empty(0, dtype=numpy.float64)),
        ('kernel_type', b'NONE'),
        ('post_transform', b'NONE'),
        ('prob_a', numpy.empty(0, dtype=numpy.float64)),
        ('prob_b', numpy.empty(0, dtype=numpy.float64)),
        ('rho', numpy.empty(0, dtype=numpy.float64)),
        ('support_vectors', numpy.empty(0, dtype=numpy.float64)),
        ('vectors_per_class', numpy.empty(0, dtype=numpy.float64)),
    ])

    def __init__(self, onnx_node, desc=None, **options):
        SVMClassifierCommon.__init__(
            self, numpy.float64, onnx_node, desc=desc,
            expected_attributes=SVMClassifierDouble.atts,
            **options)


class SVMClassifierDoubleSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl SVMClassifierDouble.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'SVMClassifierDouble')
        self.attributes = SVMClassifierDouble.atts



"""
@file Functions to creates a benchmark based on :epkg:`asv`
for many regressors and classifiers.
"""
import os
import textwrap
import hashlib
try:
    from ..onnxrt.optim.sklearn_helper import set_n_jobs
except (ValueError, ImportError):  # pragma: no cover
    from mlprodict.onnxrt.optim.sklearn_helper import set_n_jobs

# exec function does not import models but potentially
# requires all specific models used to defines scenarios
try:
    from ..onnxrt.validate.validate_scenarios import *  # pylint: disable=W0614,W0401
except (ValueError, ImportError):  # pragma: no cover
    # Skips this step if used in a benchmark.
    pass


default_asv_conf = {
    "version": 1,
    "project": "mlprodict",
    "project_url": "http://www.xavierdupre.fr/app/mlprodict/helpsphinx/index.html",
    "repo": "https://github.com/sdpython/mlprodict.git",
    "repo_subdir": "",
    "install_command": ["python -mpip install {wheel_file}"],
    "uninstall_command": ["return-code=any python -mpip uninstall -y {project}"],
    "build_command": [
        "python setup.py build",
        "PIP_NO_BUILD_ISOLATION=false python -mpip wheel --no-deps --no-index -w {build_cache_dir} {build_dir}"
    ],
    "branches": ["master"],
    "environment_type": "virtualenv",
    "install_timeout": 600,
    "show_commit_url": "https://github.com/sdpython/mlprodict/commit/",
    # "pythons": ["__PYVER__"],
    "matrix": {
        "cython": [],
        "jinja2": [],
        "joblib": [],
        "lightgbm": [],
        "mlinsights": [],
        "numpy": [],
        "onnx": ["http://localhost:8067/simple/"],
        "onnxruntime": ["http://localhost:8067/simple/"],
        "pandas": [],
        "Pillow": [],
        "pybind11": [],
        "pyquickhelper": [],
        "scipy": [],
        # "git+https://github.com/xadupre/onnxconverter-common.git@jenkins"],
        "onnxconverter-common": ["http://localhost:8067/simple/"],
        # "git+https://github.com/xadupre/sklearn-onnx.git@jenkins"],
        "skl2onnx": ["http://localhost:8067/simple/"],
        # "git+https://github.com/scikit-learn/scikit-learn.git"],
        "scikit-learn": ["http://localhost:8067/simple/"],
        "xgboost": [],
    },
    "benchmark_dir": "benches",
    "env_dir": "env",
    "results_dir": "results",
    "html_dir": "html",
}

flask_helper = """
'''
Local ASV files do no properly render in a browser,
it needs to be served through a server.
'''
import os.path
from flask import Flask, Response

app = Flask(__name__)
app.config.from_object(__name__)


def root_dir():
    return os.path.join(os.path.abspath(os.path.dirname(__file__)), "..", "html")


def get_file(filename):  # pragma: no cover
    try:
        src = os.path.join(root_dir(), filename)
        with open(src, "r", encoding="utf-8", errors="ignore") as f:
            return f.read()
    except IOError as exc:
        return str(exc)


@app.route('/', methods=['GET'])
def mainpage():
    content = get_file('index.html')
    return Response(content, mimetype="text/html")


@app.route('/', defaults={'path': ''})
@app.route('/<path:path>')
def get_resource(path):  # pragma: no cover
    mimetypes = {
        ".css": "text/css",
        ".html": "text/html",
        ".js": "application/javascript",
    }
    complete_path = os.path.join(root_dir(), path)
    ext = os.path.splitext(path)[1]
    mimetype = mimetypes.get(ext, "text/html")
    content = get_file(complete_path)
    return Response(content, mimetype=mimetype)


if __name__ == '__main__':  # pragma: no cover
    app.run(  # ssl_context=('cert.pem', 'key.pem'),
        port=8877,
        # host="",
    )
"""

pyspy_template = """
import sys
sys.path.append(r"__PATH__")
from __PYFOLD__ import __CLASSNAME__
import time
from datetime import datetime


def start():
    cl = __CLASSNAME__()
    cl.setup_cache()
    return cl


def profile0(iter, cl, runtime, N, nf, opset, dtype, optim):
    begin = time.perf_counter()
    for i in range(0, 100):
        cl.time_predict(runtime, N, nf, opset, dtype, optim)
    duration = time.perf_counter() - begin
    iter = max(100, int(25 / duration * 100)) # 25 seconds
    return iter


def setup_profile0(iter, cl, runtime, N, nf, opset, dtype, optim):
    cl.setup(runtime, N, nf, opset, dtype, optim)
    return profile0(iter, cl, runtime, N, nf, opset, dtype, optim)


def profile(iter, cl, runtime, N, nf, opset, dtype, optim):
    for i in range(iter):
        cl.time_predict(runtime, N, nf, opset, dtype, optim)
    return iter


def setup_profile(iter, cl, runtime, N, nf, opset, dtype, optim):
    cl.setup(runtime, N, nf, opset, dtype, optim)
    return profile(iter, cl, runtime, N, nf, opset, dtype, optim)


cl = start()
iter = None
print(datetime.now(), "begin")
"""


def _sklearn_subfolder(model):
    """
    Returns the list of subfolders for a model.
    """
    mod = model.__module__
    if mod is not None and mod.startswith('mlinsights'):
        return ['mlinsights', model.__name__]  # pragma: no cover
    spl = mod.split('.')
    try:
        pos = spl.index('sklearn')
    except ValueError as e:  # pragma: no cover
        raise ValueError(
            "Unable to find 'sklearn' in '{}'.".format(mod)) from e
    res = spl[pos + 1: -1]
    if len(res) == 0:
        if spl[-1] == 'sklearn':
            res = ['_externals']
        elif spl[0] == 'sklearn':
            res = spl[pos + 1:]
        else:
            raise ValueError(  # pragma: no cover
                "Unable to guess subfolder for '{}'.".format(model.__class__))
    res.append(model.__name__)
    return res


def _handle_init_files(model, flat, location, verbose, location_pyspy, fLOG):
    "Returns created, location_model, prefix_import."
    if flat:
        return ([], location, ".",
                (None if location_pyspy is None else location_pyspy))

    created = []
    subf = _sklearn_subfolder(model)
    subf = [_ for _ in subf if _[0] != '_' or _ == '_externals']
    location_model = os.path.join(location, *subf)
    prefix_import = "." * (len(subf) + 1)
    if not os.path.exists(location_model):
        os.makedirs(location_model)
        for fold in [location_model, os.path.dirname(location_model),
                     os.path.dirname(os.path.dirname(location_model))]:
            init = os.path.join(fold, '__init__.py')
            if not os.path.exists(init):
                with open(init, 'w') as _:
                    pass
                created.append(init)
                if verbose > 1 and fLOG is not None:
                    fLOG("[create_asv_benchmark] create '{}'.".format(init))
    if location_pyspy is not None:
        location_pyspy_model = os.path.join(location_pyspy, *subf)
        if not os.path.exists(location_pyspy_model):
            os.makedirs(location_pyspy_model)
    else:
        location_pyspy_model = None

    return created, location_model, prefix_import, location_pyspy_model


def _asv_class_name(model, scenario, optimisation,
                    extra, dofit, conv_options, problem,
                    shorten=True):

    def clean_str(val):
        s = str(val)
        r = ""
        for c in s:
            if c in ",-\n":
                r += "_"
                continue
            if c in ": =.+()[]{}\"'<>~":
                continue
            r += c
        for k, v in {'n_estimators': 'nest',
                     'max_iter': 'mxit'}.items():
            r = r.replace(k, v)
        return r

    def clean_str_list(val):
        if val is None:
            return ""  # pragma: no cover
        if isinstance(val, list):
            return ".".join(  # pragma: no cover
                clean_str_list(v) for v in val if v)
        return clean_str(val)

    els = ['bench', model.__name__, scenario, clean_str(problem)]
    if not dofit:
        els.append('nofit')
    if extra:
        if 'random_state' in extra and extra['random_state'] == 42:
            extra2 = extra.copy()
            del extra2['random_state']
            if extra2:
                els.append(clean_str(extra2))
        else:
            els.append(clean_str(extra))
    if optimisation:
        els.append(clean_str_list(optimisation))
    if conv_options:
        els.append(clean_str_list(conv_options))
    res = ".".join(els).replace("-", "_")

    if shorten:
        rep = {
            'ConstantKernel': 'Cst',
            'DotProduct': 'Dot',
            'Exponentiation': 'Exp',
            'ExpSineSquared': 'ExpS2',
            'GaussianProcess': 'GaussProc',
            'GaussianMixture': 'GaussMixt',
            'HistGradientBoosting': 'HGB',
            'LinearRegression': 'LinReg',
            'LogisticRegression': 'LogReg',
            'MultiOutput': 'MultOut',
            'OrthogonalMatchingPursuit': 'OrthMatchPurs',
            'PairWiseKernel': 'PW',
            'Product': 'Prod',
            'RationalQuadratic': 'RQ',
            'WhiteKernel': 'WK',
            'length_scale': 'ls',
            'periodicity': 'pcy',
        }
        for k, v in rep.items():
            res = res.replace(k, v)

        rep = {
            'Classifier': 'Clas',
            'Regressor': 'Reg',
            'KNeighbors': 'KNN',
            'NearestNeighbors': 'kNN',
            'RadiusNeighbors': 'RadNN',
        }
        for k, v in rep.items():
            res = res.replace(k, v)

        if len(res) > 70:  # shorten filename
            m = hashlib.sha256()
            m.update(res.encode('utf-8'))
            sh = m.hexdigest()
            if len(sh) > 6:
                sh = sh[:6]
            res = res[:70] + sh
    return res


def _read_patterns():
    """
    Reads the testing pattern.
    """
    # Reads the template
    patterns = {}
    for suffix in ['classifier', 'classifier_raw_scores', 'regressor', 'clustering',
                   'outlier', 'trainable_transform', 'transform',
                   'multi_classifier', 'transform_positive']:
        template_name = os.path.join(os.path.dirname(
            __file__), "template", "skl_model_%s.py" % suffix)
        if not os.path.exists(template_name):
            raise FileNotFoundError(  # pragma: no cover
                "Template '{}' was not found.".format(template_name))
        with open(template_name, "r", encoding="utf-8") as f:
            content = f.read()
        initial_content = '"""'.join(content.split('"""')[2:])
        patterns[suffix] = initial_content
    return patterns


def _select_pattern_problem(prob, patterns):
    """
    Selects a benchmark type based on the problem kind.
    """
    if '-reg' in prob:
        return patterns['regressor']
    if '-cl' in prob and '-dec' in prob:
        return patterns['classifier_raw_scores']
    if '-cl' in prob:
        return patterns['classifier']
    if 'cluster' in prob:
        return patterns['clustering']
    if 'outlier' in prob:
        return patterns['outlier']
    if 'num+y-tr' in prob:
        return patterns['trainable_transform']
    if 'num-tr-pos' in prob:
        return patterns['transform_positive']
    if 'num-tr' in prob:
        return patterns['transform']
    if 'm-label' in prob:
        return patterns['multi_classifier']
    raise ValueError(  # pragma: no cover
        "Unable to guess the right pattern for '{}'.".format(prob))


def _display_code_lines(code):
    rows = ["%03d %s" % (i + 1, line)
            for i, line in enumerate(code.split("\n"))]
    return "\n".join(rows)


def _format_dict(opts, indent):
    """
    Formats a dictionary as code.
    """
    rows = []
    for k, v in sorted(opts.items()):
        rows.append('%s=%r' % (k, v))
    content = ', '.join(rows)
    st1 = "\n".join(textwrap.wrap(content))
    return textwrap.indent(st1, prefix=' ' * indent)


def _additional_imports(model_name):
    """
    Adds additional imports for experimental models.
    """
    if model_name == 'IterativeImputer':
        return ["from sklearn.experimental import enable_iterative_imputer  # pylint: disable=W0611"]
    if model_name in ('HistGradientBoostingClassifier', 'HistGradientBoostingClassifier'):
        return ["from sklearn.experimental import enable_hist_gradient_boosting  # pylint: disable=W0611"]
    return None


def add_model_import_init(
        class_content, model, optimisation=None,
        extra=None, conv_options=None):
    """
    Modifies a template such as @see cl TemplateBenchmarkClassifier
    with code associated to the model *model*.

    @param  class_content       template (as a string)
    @param  model               model class
    @param  optimisation        model optimisation
    @param  extra               addition parameter to the constructor
    @param  conv_options        options for the conversion to ONNX
    @returm                     modified template
    """
    add_imports = []
    add_methods = []
    add_params = ["par_modelname = '%s'" % model.__name__,
                  "par_extra = %r" % extra]

    # additional methods and imports
    if optimisation is not None:
        add_imports.append(
            'from mlprodict.onnxrt.optim import onnx_optimisations')
        if optimisation == 'onnx':
            add_methods.append(textwrap.dedent('''
                def _optimize_onnx(self, onx):
                    return onnx_optimisations(onx)'''))
            add_params.append('par_optimonnx = True')
        elif isinstance(optimisation, dict):
            add_methods.append(textwrap.dedent('''
                def _optimize_onnx(self, onx):
                    return onnx_optimisations(onx, self.par_optims)'''))
            add_params.append('par_optims = {}'.format(
                _format_dict(optimisation, indent=4)))
        else:
            raise ValueError(  # pragma: no cover
                "Unable to interpret optimisation {}.".format(optimisation))

    # look for import place
    lines = class_content.split('\n')
    keep = None
    for pos, line in enumerate(lines):
        if "# Import specific to this model." in line:
            keep = pos
            break
    if keep is None:
        raise RuntimeError(  # pragma: no cover
            "Unable to locate where to insert import in\n{}\n".format(
                class_content))

    # imports
    loc_class = model.__module__
    sub = loc_class.split('.')
    if 'sklearn' not in sub:
        mod = loc_class
    else:
        skl = sub.index('sklearn')
        if skl == 0:
            if sub[-1].startswith("_"):
                mod = '.'.join(sub[skl:-1])
            else:
                mod = '.'.join(sub[skl:])
        else:
            mod = '.'.join(sub[:-1])

    exp_imports = _additional_imports(model.__name__)
    if exp_imports:
        add_imports.extend(exp_imports)
    imp_inst = (
        "try:\n    from {0} import {1}\nexcept ImportError:\n    {1} = None"
        "".format(mod, model.__name__))
    add_imports.append(imp_inst)
    add_imports.append("#  __IMPORTS__")
    lines[keep + 1] = "\n".join(add_imports)
    content = "\n".join(lines)

    # _create_model
    content = content.split('def _create_model(self):')[0].strip(' \n')
    lines = [content, "", "    def _create_model(self):"]
    if extra is not None and len(extra) > 0:
        lines.append("        return {}(".format(model.__name__))
        lines.append(_format_dict(set_n_jobs(model, extra), 12))
        lines.append("        )")
    else:
        lines.append("        return {}()".format(model.__name__))
    lines.append("")

    # methods
    for meth in add_methods:
        lines.append(textwrap.indent(meth, '    '))
        lines.append('')

    # end
    return "\n".join(lines), add_params


def find_missing_sklearn_imports(pieces):
    """
    Finds in :epkg:`scikit-learn` the missing pieces.

    @param      pieces      list of names in scikit-learn
    @return                 list of corresponding imports
    """
    res = {}
    for piece in pieces:
        mod = find_sklearn_module(piece)
        if mod not in res:
            res[mod] = []
        res[mod].append(piece)

    lines = []
    for k, v in res.items():
        lines.append("from {} import {}".format(
            k, ", ".join(sorted(v))))
    return lines


def find_sklearn_module(piece):
    """
    Finds the corresponding modulee for an element of :epkg:`scikit-learn`.

    @param      piece       name to import
    @return                 module name

    The implementation is not intelligence and should
    be improved. It is a kind of white list.
    """
    glo = globals()
    if piece in {'LinearRegression', 'LogisticRegression',
                 'SGDClassifier'}:
        import sklearn.linear_model
        glo[piece] = getattr(sklearn.linear_model, piece)
        return "sklearn.linear_model"
    if piece in {'DecisionTreeRegressor', 'DecisionTreeClassifier'}:
        import sklearn.tree
        glo[piece] = getattr(sklearn.tree, piece)
        return "sklearn.tree"
    if piece in {'ExpSineSquared', 'DotProduct', 'RationalQuadratic', 'RBF'}:
        import sklearn.gaussian_process.kernels
        glo[piece] = getattr(sklearn.gaussian_process.kernels, piece)
        return "sklearn.gaussian_process.kernels"
    if piece in {'LinearSVC', 'LinearSVR', 'NuSVR', 'SVR', 'SVC', 'NuSVC'}:  # pragma: no cover
        import sklearn.svm
        glo[piece] = getattr(sklearn.svm, piece)
        return "sklearn.svm"
    if piece in {'KMeans'}:  # pragma: no cover
        import sklearn.cluster
        glo[piece] = getattr(sklearn.cluster, piece)
        return "sklearn.cluster"
    if piece in {'OneVsRestClassifier', 'OneVsOneClassifier'}:  # pragma: no cover
        import sklearn.multiclass
        glo[piece] = getattr(sklearn.multiclass, piece)
        return "sklearn.multiclass"
    raise ValueError(  # pragma: no cover
        "Unable to find module to import for '{}'.".format(piece))



"""
@file
@brief Scenarios for validation.
"""
from sklearn.experimental import enable_hist_gradient_boosting  # pylint: disable=W0611
from sklearn import __all__ as sklearn__all__, __version__ as sklearn_version
from sklearn.cluster import KMeans
from sklearn.calibration import CalibratedClassifierCV
from sklearn.decomposition import SparseCoder, LatentDirichletAllocation
from sklearn.ensemble import (
    VotingClassifier, AdaBoostRegressor, VotingRegressor,
    ExtraTreesRegressor, ExtraTreesClassifier,
    RandomForestRegressor, RandomForestClassifier,
    HistGradientBoostingRegressor, HistGradientBoostingClassifier,
    AdaBoostClassifier, GradientBoostingClassifier, GradientBoostingRegressor,
    IsolationForest)
from sklearn.feature_extraction import DictVectorizer, FeatureHasher
from sklearn.feature_selection import (
    SelectFromModel, SelectPercentile, RFE, RFECV,
    SelectKBest, SelectFwe)
from sklearn.gaussian_process import GaussianProcessRegressor, GaussianProcessClassifier
from sklearn.gaussian_process.kernels import ExpSineSquared, DotProduct, RationalQuadratic, RBF
from sklearn.linear_model import (
    LogisticRegression, LogisticRegressionCV, SGDClassifier,
    LinearRegression, Perceptron, RidgeClassifier, RidgeClassifierCV,
    PassiveAggressiveClassifier)
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.multiclass import OneVsRestClassifier, OneVsOneClassifier, OutputCodeClassifier
from sklearn.multioutput import MultiOutputRegressor, MultiOutputClassifier, ClassifierChain, RegressorChain
from sklearn.neighbors import (
    LocalOutlierFactor, KNeighborsRegressor, KNeighborsClassifier,
    RadiusNeighborsRegressor, RadiusNeighborsClassifier)
from sklearn.neural_network import MLPClassifier
from sklearn.preprocessing import Normalizer, PowerTransformer
from sklearn.random_projection import GaussianRandomProjection, SparseRandomProjection
from sklearn.svm import SVC, NuSVC, SVR
from sklearn.tree import DecisionTreeRegressor, DecisionTreeClassifier, ExtraTreeClassifier

try:
    from sklearn.ensemble import StackingClassifier, StackingRegressor
except ImportError:  # pragma: no cover
    # new in 0.22
    StackingClassifier, StackingRegressor = None, None


def build_custom_scenarios():
    """
    Defines parameters values for some operators.

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.onnxrt.validate.validate_scenarios import build_custom_scenarios
        import pprint
        pprint.pprint(build_custom_scenarios())
    """
    options = {
        # skips
        SparseCoder: None,
        # scenarios
        AdaBoostClassifier: [
            ('default', {'n_estimators': 10},
             {'conv_options': [{AdaBoostClassifier: {'zipmap': False}}]}),
        ],
        AdaBoostRegressor: [
            ('default', {'n_estimators': 10}),
        ],
        CalibratedClassifierCV: [
            ('sgd', {
                'base_estimator': SGDClassifier(),
            }),
            ('default', {}),
        ],
        ClassifierChain: [
            ('logreg', {
                'base_estimator': LogisticRegression(solver='liblinear'),
            })
        ],
        DecisionTreeClassifier: [
            ('default', {}, {'conv_options': [
             {DecisionTreeClassifier: {'zipmap': False}}]})
        ],
        DictVectorizer: [
            ('default', {}),
        ],
        ExtraTreeClassifier: [
            ('default', {},
             {'conv_options': [{ExtraTreeClassifier: {'zipmap': False}}]}),
        ],
        ExtraTreesClassifier: [
            ('default', {'n_estimators': 10},
             {'conv_options': [{ExtraTreesClassifier: {'zipmap': False}}]}),
        ],
        ExtraTreesRegressor: [
            ('default', {'n_estimators': 10}),
        ],
        FeatureHasher: [
            ('default', {}),
        ],
        GaussianProcessClassifier: [
            ('expsine', {
                'kernel': ExpSineSquared(),
            }, {'conv_options': [{}, {GaussianProcessClassifier: {'optim': 'cdist'}}]}),
            ('dotproduct', {
                'kernel': DotProduct(),
            }, {'conv_options': [{GaussianProcessClassifier: {'optim': 'cdist'}}]}),
            ('rational', {
                'kernel': RationalQuadratic(),
            }, {'conv_options': [{GaussianProcessClassifier: {'optim': 'cdist'}}]}),
            ('rbf', {
                'kernel': RBF(),
            }, {'conv_options': [{GaussianProcessClassifier: {'optim': 'cdist'}}]}),
        ],
        GaussianProcessRegressor: [
            ('expsine', {
                'kernel': ExpSineSquared(),
                'alpha': 20.,
            }, {'conv_options': [{GaussianProcessRegressor: {'optim': 'cdist'}}]}),
            ('dotproduct', {
                'kernel': DotProduct(),
                'alpha': 100.,
            }, {'conv_options': [{}, {GaussianProcessRegressor: {'optim': 'cdist'}}]}),
            ('rational', {
                'kernel': RationalQuadratic(),
                'alpha': 100.,
            }, {'conv_options': [{GaussianProcessRegressor: {'optim': 'cdist'}}]}),
            ('rbf', {
                'kernel': RBF(),
                'alpha': 100.,
            }, {'conv_options': [{GaussianProcessRegressor: {'optim': 'cdist'}}]}),
        ],
        GaussianRandomProjection: [
            ('eps95', {'eps': 0.95}),
        ],
        GradientBoostingClassifier: [
            ('default', {'n_estimators': 200},
             {'conv_options': [{GradientBoostingClassifier: {'zipmap': False}}]}),
        ],
        GradientBoostingRegressor: [
            ('default', {'n_estimators': 200}),
        ],
        GridSearchCV: [
            ('cl', {
                'estimator': LogisticRegression(solver='liblinear'),
                'param_grid': {'fit_intercept': [False, True]}},
             {'conv_options': [{GridSearchCV: {'zipmap': False}}],
              'subset_problems': ['b-cl', 'm-cl', '~b-cl-64']}),
            ('reg', {
                'estimator': LinearRegression(),
                'param_grid': {'fit_intercept': [False, True]},
            }, ['b-reg', 'm-reg', '~b-reg-64']),
            ('reg', {
                'estimator': KMeans(),
                'param_grid': {'n_clusters': [2, 3]},
            }, ['cluster']),
        ],
        HistGradientBoostingClassifier: [
            ('default', {'max_iter': 100},
             {'conv_options': [{HistGradientBoostingClassifier: {'zipmap': False}}]}),
        ],
        HistGradientBoostingRegressor: [
            ('default', {'max_iter': 100}),
        ],
        IsolationForest: [
            ('default', {'n_estimators': 10}),
        ],
        KNeighborsClassifier: [
            ('default_k3', {'algorithm': 'brute', 'n_neighbors': 3},
             {'conv_options': [{KNeighborsClassifier: {'optim': 'cdist', 'zipmap': False}}]}),
            ('weights_k3', {'algorithm': 'brute',
                            'weights': 'distance', 'n_neighbors': 3},
             {'conv_options': [{KNeighborsClassifier: {'optim': 'cdist', 'zipmap': False}}]}),
        ],
        KNeighborsRegressor: [
            ('default_k3', {'algorithm': 'brute', 'n_neighbors': 3},
             {'conv_options': [{KNeighborsRegressor: {'optim': 'cdist'}}]}),
            ('weights_k3', {'algorithm': 'brute',
                            'weights': 'distance', 'n_neighbors': 3},
             {'conv_options': [{KNeighborsRegressor: {'optim': 'cdist'}}]}),
        ],
        LatentDirichletAllocation: [
            ('default', {'n_components': 2}),
        ],
        LocalOutlierFactor: [
            ('novelty', {'novelty': True}),
        ],
        LogisticRegression: [
            ('liblinear', {'solver': 'liblinear', },
             {'optim': [None, 'onnx'],
              'conv_options': [{}, {LogisticRegression: {'zipmap': False}}],
              'subset_problems': ['b-cl', '~b-cl-64', 'm-cl']}),
            ('liblinear-dec',
             {'solver': 'liblinear', },
             {'conv_options': [{LogisticRegression: {'raw_scores': True, 'zipmap': False}}],
              'subset_problems': ['~b-cl-dec', '~m-cl-dec']}),
        ],
        LogisticRegressionCV: [
            ('default', {},
             {'conv_options': [{LogisticRegressionCV: {'zipmap': False}}]}),
        ],
        MLPClassifier: [
            ('default', {}, {'conv_options': [
             {MLPClassifier: {'zipmap': False}}]}),
        ],
        MultiOutputClassifier: [
            ('logreg', {
                'estimator': LogisticRegression(solver='liblinear')},
             {'conv_options': [{MultiOutputClassifier: {'zipmap': False}}]},)
        ],
        MultiOutputRegressor: [
            ('linreg', {
                'estimator': LinearRegression(),
            })
        ],
        Normalizer: [
            ('l2', {'norm': 'l2', }),
            ('l1', {'norm': 'l1', }),
            ('max', {'norm': 'max', }),
        ],
        NuSVC: [
            ('prob', {
                'probability': True,
            }),
        ],
        OneVsOneClassifier: [
            ('logreg', {'estimator': LogisticRegression(solver='liblinear')},
             {'conv_options': [{OneVsOneClassifier: {'zipmap': False}}]})
        ],
        OneVsRestClassifier: [
            ('logreg', {'estimator': LogisticRegression(solver='liblinear')},
             {'conv_options': [{OneVsOneClassifier: {'zipmap': False}}]})
        ],
        OutputCodeClassifier: [
            ('logreg', {'estimator': LogisticRegression(solver='liblinear')},
             {'conv_options': [{OneVsOneClassifier: {'zipmap': False}}]})
        ],
        PassiveAggressiveClassifier: [
            ('logreg', {}, {'conv_options': [
             {PassiveAggressiveClassifier: {'zipmap': False}}]})
        ],
        Perceptron: [
            ('logreg', {}, {'conv_options': [{Perceptron: {'zipmap': False}}]})
        ],
        PowerTransformer: [
            ('yeo-johnson', {'method': 'yeo-johnson'}),
            ('box-cox', {'method': 'box-cox'}),
        ],
        RadiusNeighborsClassifier: [
            ('default_k3', {'algorithm': 'brute', 'n_neighbors': 3},
             {'conv_options': [{RadiusNeighborsClassifier: {'optim': 'cdist', 'zipmap': False}}]}),
            ('weights_k3', {'algorithm': 'brute',
                            'weights': 'distance', 'n_neighbors': 3},
             {'conv_options': [{RadiusNeighborsClassifier: {'optim': 'cdist', 'zipmap': False}}]}),
        ],
        RadiusNeighborsRegressor: [
            ('default_k3', {'algorithm': 'brute', 'n_neighbors': 3},
             {'conv_options': [{}, {RadiusNeighborsRegressor: {'optim': 'cdist'}}]}),
            ('weights_k3', {'algorithm': 'brute',
                            'weights': 'distance', 'n_neighbors': 3},
             {'conv_options': [{RadiusNeighborsRegressor: {'optim': 'cdist'}}]}),
        ],
        RandomForestClassifier: [
            ('default', {'n_estimators': 100},
             {'conv_options': [{RandomForestClassifier: {'zipmap': False}}]}),
        ],
        RandomForestRegressor: [
            ('default', {'n_estimators': 100}),
        ],
        RandomizedSearchCV: [
            ('cl', {
                'estimator': LogisticRegression(solver='liblinear'),
                'param_distributions': {'fit_intercept': [False, True]},
            }),
            ('reg', {
                'estimator': LinearRegression(),
                'param_distributions': {'fit_intercept': [False, True]},
            }),
        ],
        RegressorChain: [
            ('linreg', {
                'base_estimator': LinearRegression(),
            })
        ],
        RidgeClassifier: [
            ('default', {},
             {'conv_options': [{RidgeClassifier: {'zipmap': False}}]}),
        ],
        RidgeClassifierCV: [
            ('default', {},
             {'conv_options': [{RidgeClassifierCV: {'zipmap': False}}]}),
        ],
        RFE: [
            ('reg', {
                'estimator': LinearRegression(),
            })
        ],
        RFECV: [
            ('reg', {
                'estimator': LinearRegression(),
            })
        ],
        SelectFromModel: [
            ('rf', {
                'estimator': DecisionTreeRegressor(),
            }),
        ],
        SelectFwe: [
            ('alpha100', {
                'alpha': 100.0,
            }),
        ],
        SelectKBest: [
            ('k2', {
                'k': 2,
            }),
        ],
        SelectPercentile: [
            ('p50', {
                'percentile': 50,
            }),
        ],
        SGDClassifier: [
            ('log', {'loss': 'log'},
             {'conv_options': [{SGDClassifier: {'zipmap': False}}]}),
        ],
        SparseRandomProjection: [
            ('eps95', {'eps': 0.95}),
        ],
        SVC: [
            ('linear', {'probability': True, 'kernel': 'linear'},
             {'conv_options': [{SVC: {'zipmap': False}}]}),
            ('poly', {'probability': True, 'kernel': 'poly'},
             {'conv_options': [{SVC: {'zipmap': False}}]}),
            ('rbf', {'probability': True, 'kernel': 'rbf'},
             {'conv_options': [{SVC: {'zipmap': False}}]}),
            ('sigmoid', {'probability': True, 'kernel': 'sigmoid'},
             {'conv_options': [{SVC: {'zipmap': False}}]}),
        ],
        SVR: [
            ('linear', {'kernel': 'linear'}),
            ('poly', {'kernel': 'poly'}),
            ('rbf', {'kernel': 'rbf'}),
            ('sigmoid', {'kernel': 'sigmoid'}),
        ],
        VotingClassifier: [
            ('logreg-noflatten', {
                'voting': 'soft',
                'flatten_transform': False,
                'estimators': [
                    ('lr1', LogisticRegression(
                        solver='liblinear', fit_intercept=True)),
                    ('lr2', LogisticRegression(
                        solver='liblinear', fit_intercept=False)),
                ],
            }, {'conv_options': [{VotingClassifier: {'zipmap': False}}]})
        ],
        VotingRegressor: [
            ('linreg', {
                'estimators': [
                    ('lr1', LinearRegression()),
                    ('lr2', LinearRegression(fit_intercept=False)),
                ],
            })
        ],
    }
    if StackingClassifier is not None and StackingRegressor is not None:
        options.update({
            StackingClassifier: [
                ('logreg', {
                    'estimators': [
                        ('lr1', LogisticRegression(solver='liblinear')),
                        ('lr2', LogisticRegression(
                            solver='liblinear', fit_intercept=False)),
                    ],
                }, {'conv_options': [{StackingClassifier: {'zipmap': False}}]})
            ],
            StackingRegressor: [
                ('linreg', {
                    'estimators': [
                        ('lr1', LinearRegression()),
                        ('lr2', LinearRegression(fit_intercept=False)),
                    ],
                })
            ],
        })
    return options


def interpret_options_from_string(st):
    """
    Converts a string into a dictionary.

    @param      st      string
    @return             evaluated object
    """
    if isinstance(st, dict):
        return st  # pragma: no cover
    value = eval(st)  # pylint: disable=W0123
    return value


_extra_parameters = build_custom_scenarios()



"""
@file
@brief Validates runtime for many :scikit-learn: operators.
The submodule relies on :epkg:`onnxconverter_common`,
:epkg:`sklearn-onnx`.
"""
import numpy
from skl2onnx.common.data_types import (
    FloatTensorType, DoubleTensorType)


text_alpha_num = [
    ('zero', 0),
    ('one', 1),
    ('two', 2),
    ('three', 3),
    ('four', 4),
    ('five', 5),
    ('six', 6),
    ('seven', 7),
    ('eight', 8),
    ('nine', 9),
    ('dix', 10),
    ('eleven', 11),
    ('twelve', 12),
    ('thirteen', 13),
    ('fourteen', 14),
    ('fifteen', 15),
    ('sixteen', 16),
    ('seventeen', 17),
    ('eighteen', 18),
    ('nineteen', 19),
    ('twenty', 20),
    ('twenty one', 21),
    ('twenty two', 22),
    ('twenty three', 23),
    ('twenty four', 24),
    ('twenty five', 25),
    ('twenty six', 26),
    ('twenty seven', 27),
    ('twenty eight', 28),
    ('twenty nine', 29),
]


def _guess_noshape(obj, shape):
    if isinstance(obj, numpy.ndarray):
        if obj.dtype == numpy.float32:
            return FloatTensorType(shape)  # pragma: no cover
        if obj.dtype == numpy.float64:
            return DoubleTensorType(shape)
        raise NotImplementedError(  # pragma: no cover
            "Unable to process object(1) [{}].".format(obj))
    raise NotImplementedError(  # pragma: no cover
        "Unable to process object(2) [{}].".format(obj))


def _noshapevar(fct):

    def process_itt(itt, Xort):
        if isinstance(itt, tuple):
            return (process_itt(itt[0], Xort), itt[1])

        # name = "V%s_" % str(id(Xort))[:5]
        new_itt = []
        for a, b in itt:
            # shape = [name + str(i) for s in b.shape]
            shape = [None for s in b.shape]
            new_itt.append((a, _guess_noshape(b, shape)))
        return new_itt

    def new_fct(**kwargs):
        X, y, itt, meth, mo, Xort = fct(**kwargs)
        new_itt = process_itt(itt, Xort)
        return X, y, new_itt, meth, mo, Xort
    return new_fct


def _1d_problem(fct):

    def new_fct(**kwargs):
        n_features = kwargs.get('n_features', None)
        if n_features not in (None, 1):
            raise RuntimeError(  # pragma: no cover
                "Misconfiguration: the number of features must not be "
                "specified for a 1D problem.")
        X, y, itt, meth, mo, Xort = fct(**kwargs)
        new_itt = itt  # process_itt(itt, Xort)
        X = X[:, 0]
        return X, y, new_itt, meth, mo, Xort
    return new_fct



"""
@file
@brief Shortcuts to *asv_benchmark*.
"""

from .asv_exports import export_asv_json
from .common_asv_skl import (
    _CommonAsvSklBenchmarkClassifier,
    _CommonAsvSklBenchmarkClassifierRawScore,
    _CommonAsvSklBenchmarkClustering,
    _CommonAsvSklBenchmarkMultiClassifier,
    _CommonAsvSklBenchmarkOutlier,
    _CommonAsvSklBenchmarkRegressor,
    _CommonAsvSklBenchmarkTrainableTransform,
    _CommonAsvSklBenchmarkTransform,
    _CommonAsvSklBenchmarkTransformPositive,
)
from .create_asv import create_asv_benchmark



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from scipy.special import expit as logistic_sigmoid  # pylint: disable=E0611
from ._op import OpRunUnaryNum


class Sigmoid(OpRunUnaryNum):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnaryNum.__init__(self, onnx_node, desc=desc,
                               **options)

    def _run(self, x):  # pylint: disable=W0221
        y = logistic_sigmoid(x)
        return (y, )

    def to_python(self, inputs):
        return ("from scipy.special import expit",
                "return expit(%s)" % inputs[0])



"""
@file
@brief Creates missing types in onnxconverter-common.

.. versionadded:: 0.6
"""
from onnx import onnx_pb as onnx_proto  # pylint: disable=W0611
from skl2onnx.common.data_types import (  # pylint: disable=W0611
    TensorType, FloatTensorType, Int64TensorType, DoubleTensorType,
    StringTensorType, Int32TensorType, BooleanTensorType,
    UInt8TensorType)
from skl2onnx.common.data_types import (  # pylint: disable=W0611
    Int16TensorType, Int8TensorType, UInt16TensorType,
    UInt32TensorType, UInt64TensorType, Float16TensorType)
